//
/*! libresignin
*/
// Copyright 2022, Jesús Pérez Lorenzo
//

use reqwest_middleware::{ClientBuilder, ClientWithMiddleware};
use reqwest_retry::{RetryTransientMiddleware, policies::ExponentialBackoff};
use reqwest_tracing::TracingMiddleware;
use serde::{Deserialize}; //, Serialize};
use axum::{
	response::Html,
	routing::get,
	// routing::post,
	Router,
	//Json,
	response::IntoResponse,
	extract::{Extension,Path,Query},
	//http::{Request, header::HeaderMap, Method,StatusCode},
	http::{
		header::{HeaderMap,HeaderName,HeaderValue},
		StatusCode,
  },
  // body::{Bytes, Body},
};
// use serde::{Deserialize, Serialize};
use crate::defs::{DataDBs}; 
use crate::utils::reqenv::ReqEnv;
use std::{
	collections::HashMap,
  sync::Arc,
};

use reqwest_cookie_store::CookieStoreMutex;

use crate::kratos::{
	get_root_url,
	get_flow_csrf,
	// load_idschema,
 	// KratosUser,
	// KratosItemProps,
	// KratosTraitItem,
  // JsonMap,
	load_traits,
	set_kratos_user,
	get_kratos_user,
	set_kratos_login,
	on_kratos_user,
	logout_kratos_user,
//	KratosSessionResponse,
//	KratosReqId,
// 	KratosResponse,
};

// use cookie::{Cookie};
// use tower_cookies::{Cookies};

pub async fn run(client: ClientWithMiddleware) {
    client
        .get("https://truelayer.com")
        .header("foo", "bar")
        .send()
        .await
        .unwrap();
}

// pub async fn registration_handler(cookies: Cookies,header: HeaderMap, Extension(dbs): Extension<DataDBs>, Extension(req_cli): Extension<Arc<CookieStoreMutex>>) -> Html<&'static str> {
pub async fn registration_handler(header: HeaderMap, Extension(dbs): Extension<DataDBs>,Extension(req_cli): Extension<Arc<CookieStoreMutex>>) ->  impl IntoResponse {
  let reqenv = ReqEnv::new(dbs.app, dbs.auth, header, "get", "/login", "html", "rt");
	let retry_policy = ExponentialBackoff::builder().build_with_max_retries(3);
  let client = ClientBuilder::new(reqwest::Client::new())
     // Trace HTTP requests. See the tracing crate to make use of these traces.
    .with(TracingMiddleware)
    // Retry failed requests.
    .with(RetryTransientMiddleware::new_with_policy(retry_policy))
    .build();

  //  run(client).await;

//  let reqwest_client = Client::builder().build().unwrap();
//     let client = ClientBuilder::new(reqwest_client)
//         .with(LoggingMiddleware)
//         .build();
//     let resp = client.get("https://truelayer.com").send().await.unwrap();
//     println!("TrueLayer page HTML: {}", resp.text().await.unwrap());


	// // dbg!("{:#?}",reqenv.websrvr());	
	let root_url = get_root_url(reqenv.websrvr().signin.protocol,reqenv.websrvr().signin.root,reqenv.websrvr().signin.port);
	let flow = String::from("self-service/registration");
	// //let flow = String::from("login");
	let query = String::from("/api?refresh=false&aal=&return_to=");
  let traits = load_traits(reqenv.websrvr().signin.idschema_path);
  // // dbg!("{:#?}",&traits);	
	let mut user_data = HashMap::new();
	user_data.insert("name.first".to_string(),"Jesús".to_string());
	user_data.insert("name.last".to_string(),"Pérez".to_string());
	user_data.insert("username".to_string(),"jesuspl".to_string());
	user_data.insert("password".to_string(),"19Ting22".to_string());
	user_data.insert("email".to_string(),"jpl@jesusperez.pro".to_string());
	let user = set_kratos_user("traits",user_data,traits);
  // // println!("{}",&user);	
	let url=format!("{}/{}{}",&root_url,flow,query);
	let html:String;
	// match get_flow_csrf(reqenv.websrvr().signin,&req_cli,url.as_str()).await {
	match get_flow_csrf("",reqenv.websrvr().signin,&req_cli,"","",url.as_str()).await {
		Ok(reqid) => {
			println!("flow: {}",&reqid.flow);
			println!("csrf_token: {}",&reqid.csrf);
			println!("cookie: {}",&reqid.cookie);
			let query_id = format!("?flow={}",&reqid.flow);
			let reg_url=format!("{}/{}{}",&root_url,flow,query_id);
			match on_kratos_user("",reqenv.websrvr().signin,&req_cli,&reqid.csrf,&reqid.cookie,&reg_url,user).await {
				Ok(res) => {
					println!("Result {:?}",res.status);
					html = format!(
					r#"
					 <h1>Register</h1>
				   <a href='{}/login'>login</a>
					 <div>Status: {} </div>
			   	"#,&reqenv.public_url(),res.status);
				},
				Err(e) => {
	        eprintln!("{}",e);
					html = format!(
					r#"
					 <h1>Login</h1>
					 <div>Error: {} </div>
			   	"#,e);
				}
			}
		},
		Err(e) => {
	    eprintln!("{}",e);
			html = format!(
			r#"
				<h1>Login</h1>
				<div>Error: {} </div>
			"#,e);
		},
	}
	Html(html)
}
pub async fn login_handler(header: HeaderMap, Extension(dbs): Extension<DataDBs>,Extension(req_cli): Extension<Arc<CookieStoreMutex>>) -> impl IntoResponse {
  let reqenv = ReqEnv::new(dbs.app.to_owned(), dbs.auth.to_owned(), header.to_owned(), "get", "/login", "html", "rt");
	// let retry_policy = ExponentialBackoff::builder().build_with_max_retries(3);
  // let client = ClientBuilder::new(reqwest::Client::new())
  //    // Trace HTTP requests. See the tracing crate to make use of these traces.
  //   .with(TracingMiddleware)
  //   // Retry failed requests.
  //   .with(RetryTransientMiddleware::new_with_policy(retry_policy))
  //   .build();

  //  run(client).await;

//  let reqwest_client = Client::builder().build().unwrap();
//     let client = ClientBuilder::new(reqwest_client)
//         .with(LoggingMiddleware)
//         .build();
//     let resp = client.get("https://truelayer.com").send().await.unwrap();
//     println!("TrueLayer page HTML: {}", resp.text().await.unwrap());


	// // dbg!("{:#?}",reqenv.websrvr());	
	let flow = String::from("self-service/login");
	let query = String::from("/api?refresh=false&aal=&return_to=");
	let root_url = get_root_url(reqenv.websrvr().signin.protocol,reqenv.websrvr().signin.root,reqenv.websrvr().signin.port);
	let mut headers = HeaderMap::new();
	let ret_status: StatusCode;
	let mut user_data = HashMap::new();
	user_data.insert("password".to_string(),"19Ting22".to_string());
	user_data.insert("password_identifier".to_string(),"jesuspl".to_string());
	let user = set_kratos_login(user_data);
	let url=format!("{}/{}{}",&root_url,flow,query);
	// match get_flow_csrf(reqenv.websrvr().signin,&req_cli,url.as_str()).await {
	let html:String;
	match get_flow_csrf("",reqenv.websrvr().signin,&req_cli,"","",url.as_str()).await {
		Ok(reqid) => {
			println!("flow: {}",&reqid.flow);
			println!("csrf_token: {}",&reqid.csrf);
			println!("cookie: {}",&reqid.cookie);
			let query_id = format!("?flow={}",&reqid.flow);
			let reg_url=format!("{}/{}{}",&root_url,flow,query_id);
			match on_kratos_user("",reqenv.websrvr().signin,&req_cli,&reqid.csrf,&reqid.cookie,&reg_url,user).await {
				Ok(res) => {
					println!("Result {:?}",res.status);
					//    res.session_resp.session.identity.id,
					// 	 res.session_resp.session_token);
          // return html.to_owned();
					if res.status == StatusCode::OK {
						headers.insert(
							HeaderName::from_static("x-authz"),
							HeaderValue::from_str(&format!("{}:tk:{}",res.session_resp.session.identity.id, res.session_resp.session_token)).unwrap_or(HeaderValue::from_static("")),
						);
						ret_status = StatusCode::OK;
						html = format!(
						r#"
						<h1>Login</h1>
						<div><a href='{}/whoami/{}?token={}'>who am i ?</a></div>
						<div><a href='{}/logout/{}?token={}'>logout</a></div>
						<div>Status: {} </div>
						"#,
						&reqenv.public_url(),res.session_resp.session.identity.id, res.session_resp.session_token,
						&reqenv.public_url(),res.session_resp.session.identity.id, res.session_resp.session_token,
						res.status);	
					} else {
						ret_status = StatusCode::OK;
						html = format!(
						r#"
						<h1>Login</h1>
						<div>Status: {} </div>
					  <div><a href='{}/registration'>Registration</a></div>
						"#,res.status,&reqenv.public_url());	
					}
				},
				Err(e) => {
	        eprintln!("{}",e);
				  // ret_status = StatusCode::FORBIDDEN;
				  ret_status = StatusCode::NOT_FOUND;
					html = format!(
					r#"
					 <h1>Login</h1>
					 <div><a href='{}/registration'>Registration</a></div>
					 <div>Error: {} </div>
			   	"#,&reqenv.public_url(),e);
				}
			}
		},
		Err(e) => {
	    eprintln!("{}",e);
		  ret_status = StatusCode::BAD_REQUEST;
		  // ret_status = StatusCode::UNAUTHORIZED;
			html = format!(
			r#"
				<h1>Login</h1>
				<div>Error: No Login Found </div>
			"#);
		},
	};
  (ret_status, headers, Html(html))
	// (StatusCode, HeaderMap, &'static str)
	//Html(html)
}
pub async fn recovery_handler() -> Html<&'static str> {
    Html("<h1>Hello, World!</h1>")
}
pub async fn whoami(id: String, token: String,
	header: HeaderMap, dbs: DataDBs,req_cli: Arc<CookieStoreMutex>) -> String {
	let reqenv = ReqEnv::new(dbs.app, dbs.auth, header, "get", "/login", "html", "rt");
	let flow = String::from("self-service/login");
	let query = String::from("/api?refresh=false&aal=&return_to=");
	let root_url = get_root_url(reqenv.websrvr().signin.protocol,reqenv.websrvr().signin.root,reqenv.websrvr().signin.port);
	let url=format!("{}/{}{}",&root_url,flow,query);
	let html:String;
	// match get_flow_csrf(reqenv.websrvr().signin,&req_cli,url.as_str()).await {
	match get_flow_csrf("",reqenv.websrvr().signin,&req_cli,"","",url.as_str()).await {
		Ok(reqid) => {
			println!("flow: {}",&reqid.flow);
			println!("csrf_token: {}",&reqid.csrf);
			println!("cookie: {}",&reqid.cookie);
	    let flow_check = String::from("sessions/whoami");
			let query_id = format!("?flow={}",&reqid.flow);
			let reg_url=format!("{}/{}{}",&root_url,flow_check,query_id);
			// match on_kratos_user(reqenv.websrvr().signin,&reqid.csrf,&id,&reg_url,user).await {
			match get_kratos_user(&id,&token,reqenv.websrvr().signin,&req_cli,&reqid.csrf,&reqid.cookie,&reg_url).await {
				Ok(res) => {
					println!("Result {:?}",res.status);
					if res.status == StatusCode::OK {
						html = format!(
						r#"
						<h1>whoami</h1>
						<div>Status: {}</div>
						<div><a href='{}/logout/{}?token={}'>logout</a></div>
						"#,res.status,
							&reqenv.public_url(),id,token)
					} else {
						html = format!(
						r#"
						<h1>whoami</h1>
						<div>Status: {}</div>
						<div><a href='{}/login'>login</a></div>
						"#,res.status, &reqenv.public_url())
					}
				},
				Err(e) => {
	        eprintln!("{}",e);
					html = format!(
					r#"
						<h1>whoami</h1>
					  <div><a href='{}/login'>Login</a></div>
						<div>Error: {} </div>
					"#,&reqenv.public_url(),e);
				}
			}
		},
		Err(e) => {
	    eprintln!("{}",e);
			html = format!(
			r#"
				<h1>whoami</h1>
				<div>Error: {} </div>
			"#,e);
		},
	};
	html
}
#[derive(Deserialize)]
pub struct QueryToken {
	pub token: String,
}
pub async fn whoami_handler(Path(id): Path<String>, query: Query<QueryToken>,
	header: HeaderMap, Extension(dbs): Extension<DataDBs>,Extension(req_cli): Extension<Arc<CookieStoreMutex>>) -> impl IntoResponse {
  Html(whoami(id,query.token.to_string(),header,dbs,req_cli).await)
}
 pub async fn logout_handler(Path(id): Path<String>, urlquery: Query<QueryToken>, header: HeaderMap, Extension(dbs): Extension<DataDBs>,Extension(req_cli): Extension<Arc<CookieStoreMutex>>) -> impl IntoResponse {
	let reqenv = ReqEnv::new(dbs.app, dbs.auth, header, "get", "/login", "html", "rt");
	let flow = String::from("self-service/login");
	let query = format!("/api?refresh=false&aal=&return_to=&query={}",urlquery.token.to_string());
	let root_url = get_root_url(reqenv.websrvr().signin.protocol,reqenv.websrvr().signin.root,reqenv.websrvr().signin.port);
	let url=format!("{}/{}{}",&root_url,flow,query);
	let html:String;
	// match get_flow_csrf(reqenv.websrvr().signin,&req_cli,url.as_str()).await {
	match get_flow_csrf("",reqenv.websrvr().signin,&req_cli,"","",url.as_str()).await {
		Ok(reqid) => {
		println!("flow: {}",&reqid.flow);
		println!("csrf_token: {}",&reqid.csrf);
		println!("cookie: {}",&reqid.cookie);
	  let flow_logout = String::from("self-service/logout");
		let query_id = format!("/api?{}&flow={}",&urlquery.token,&reqid.flow);
		let reg_url=format!("{}/{}{}",&root_url,flow_logout,query_id);
		// match on_kratos_user(reqenv.websrvr().signin,&reqid.csrf,&id,&reg_url,user).await {
		match logout_kratos_user(&id,&urlquery.token.to_string(),reqenv.websrvr().signin,&req_cli,&reqid.csrf,&reqid.cookie,&reg_url).await {
			Ok(res) => {
				println!("Result {:?}",res.status);
				html = format!(
				r#"
				<h1>Logout</h1>
				<a href='{}/login'>login</a>
				<div>Status: {}</div>
				"#,reqenv.public_url(),res.status);
			},
			Err(e) => {
				eprintln!("{}",e);
				html = format!(
				r#"
					<h1>Logout</h1>
					<div>Error: {} </div>
				"#,e);
			}
		};
	},
	Err(e) => {
			eprintln!("{}",e);
			html = format!(
			r#"
				<h1>Logout</h1>
				<div>Error: {} </div>
			"#,e);
		},
	};
	Html(html)
}
pub async fn verification_handler() -> Html<&'static str> {
   Html("<h1>Hello, World!</h1>")
}
pub async fn settings_handler() -> Html<&'static str> {
    Html("<h1>Hello, World!</h1>")
}
pub async fn welcome_handler() -> Html<&'static str> {
    Html("<h1>Hello, World!</h1>")
}

pub fn def_handlers(web_router: Router) -> Router {
  web_router
  .route("/login", get(login_handler))
  .route("/registration", get(registration_handler))
  .route("/recovery", get(recovery_handler))
  .route("/whoami/:id", get(whoami_handler))
  .route("/logout/:id", get(logout_handler))
  .route("/verification", get(verification_handler))
  .route("/settings", get(settings_handler))
  .route("/welcome", get(welcome_handler))
}