//
/*! libresignin
*/
// Copyright 2022, Jesús Pérez Lorenzo
//

// use axum::{
// 	response::Html,
// 	routing::get,
// 	routing::post,
// 	Router,
// 	extract::{self,Extension,Path,Query},
// 	http::{Request, header::HeaderMap, Method},
//   body::{Bytes, Body},
// };
use std::collections::HashMap;
use serde::{Deserialize}; //, Serialize};
use reqwest::header::{
	HeaderMap,
	HeaderValue,
	USER_AGENT, AUTHORIZATION, SET_COOKIE, COOKIE,	ACCEPT, CONTENT_TYPE
};
use reqwest::StatusCode;
//use reqwest::cookie::Cookie;
// use tower_cookies::{Cookie, Cookies};
use anyhow::{anyhow,Result};
use reqwest_cookie_store::CookieStoreMutex;
use app_env::config::SigninServer; 
use std::sync::Arc;

const PASSWORD_FIELD: &str = "password";

// use crate::defs::{DataDBs}; 
// use crate::utils::reqenv::ReqEnv;
/*
fn parse_url(text: &str, lang: &str) -> reqwest::Url {
    let mut url = reqwest::Url::parse("https://translate.google.com/translate_tts?ie=UTF-8&total=1&idx=0&client=tw-ob").unwrap();
    url.query_pairs_mut()
        .append_pair("tl", lang)
        .append_pair("q", text)
        .append_pair("textlen", &text.len().to_string())
        .finish();
    url
}


fn get_url_flow(base: String, flow: String, query: String ) -> String {
	format!("{}/self-service/{}/browser{}",base,flow,query)
}
*/
fn default_empty() -> String {
  "/".to_string()
}
fn default_array_empty() -> Vec<String> {
  Vec::new()
}
fn default_false() -> bool {
	false
}
fn default_true() -> bool {
	true
}

// TODO load Traits from json file  "identity.schema.json" .properties.traits 
// What about additionalProperties
pub type JsonMap = serde_json::Map<String, serde_json::Value>;

#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosItemProps {
	pub title: String,
	#[serde(alias = "type")]
	pub typ: String,
}

#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosTraitItem {
	#[serde(alias = "type")]
  pub typ: String,
	pub properties: Option<HashMap::<String,KratosItemProps>>,
} 

#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosUser {
	pub traits: HashMap::<String,KratosTraitItem>,
	pub password: String,
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosMetaLabel {
  id: String,
  text: String,
	#[serde(alias = "type")]
  typ: String,
  context: String, // {}
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosMeta {
	label: KratosMetaLabel,
}
fn default_kratos_meta() -> KratosMeta {
	KratosMeta::default()
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosAttrib {
	name: String,
	#[serde(alias = "type")]
  typ: String,
	#[serde(default = "default_empty")]
  value: String,
	#[serde(default = "default_false")]
  required: bool,
	#[serde(default = "default_true")]
  disabled: bool,
  node_type: String,
	#[serde(default = "default_array_empty")]
	messages: Vec<String>,
	#[serde(default = "default_kratos_meta")]
	meta: KratosMeta, 
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosNode {
	#[serde(alias = "type")]
	typ: String,
	group: String,
	attributes: KratosAttrib,
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosUi {
  action: String,
  method: String,
  nodes: Vec<KratosNode>,
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosHeader {
	#[serde(alias = "set-cookie")]
  cookie: String,
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosIdentity {
	pub created_at: String,
	pub id: String,
	// pub recovery_addresses: Vec<serde_json::Value>,
	pub schema_id: String,
	pub schema_url: String,
	pub state: String,
	pub state_changed_at: String,
	//pub traits: Vec<serde_json::Value>,
	pub updated_at: String,
	//pub verifiable_addresses: Vec<serde_json::Value>,
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosSession {
	pub id: String,
	pub active: bool,
	pub expires_at: String,
	pub authenticated_at: String,
	pub issued_at: String,
	pub identity: KratosIdentity,
	// pub authentication_methods: Vec<serde_json::Value>,
}
fn default_default_session() -> KratosSession {
	KratosSession::default()
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosSessionResponse {
	#[serde(default = "default_empty")]
	pub session_token: String,
	#[serde(default = "default_default_session")]
	pub session: KratosSession,
}

#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosResponse {
	id: String,
	#[serde(alias = "type")]
  typ: String,
  expires_at: String,
  issued_at: String,
  request_url: String,
  ui: KratosUi,
}
#[derive(Debug, Deserialize,Default)]
#[allow(dead_code)]
pub struct KratosReqId {
	pub flow: String,
	pub csrf: String,
	pub cookie: String,
}
#[derive(Debug,Default)]
#[allow(dead_code)]
pub struct OnKratosReq {
	pub status: StatusCode,
	pub cookie: String,
	pub session_resp: KratosSessionResponse,
}

fn load_idschema(path: String) -> JsonMap {
  let str_data = std::fs::read_to_string(&path)
    .unwrap_or_else(|e| {
    	eprintln!("Error read idschema {}: {}",&path,e);
   		 String::from("")
  });
  if str_data.is_empty() {
		JsonMap::default()
	} else {
		serde_json::from_str(&str_data).unwrap_or_else(|e| {
			eprintln!("Error loading idschema {}: {}",&path,e);
		  JsonMap::default()
		})
	}
}

fn get_item_props(val: &serde_json::Value) -> HashMap::<String,KratosItemProps>  {
  let mut props: HashMap::<String,KratosItemProps> = HashMap::new();
  if let Some(data_obj)= val["properties"].as_object() { 
		for (ky,val) in data_obj {
			props.insert(ky.to_owned(), KratosItemProps {
				title: format!("{}",val["title"]).replace('"',""),
				typ: format!("{}",val["type"]).replace('"',""),
			});
		}
	}
	props
}
pub fn load_traits(path: String) -> HashMap::<String,KratosTraitItem> {
	let idschema: JsonMap = load_idschema(path);
	// dbg!("{:#?}",&idschema["properties"]["traits"]["properties"]);
	let mut traits: HashMap::<String,KratosTraitItem> = HashMap::new();
  if let Some(data_obj)= idschema["properties"]["traits"]["properties"].as_object() {
		for (ky,val) in data_obj {
			let value = format!("{}",val["type"]).replace('"',"");
			match value.as_str() {
				"object" => {
					traits.insert(ky.to_owned(),KratosTraitItem {
						typ: value,
						properties: Some(get_item_props(val)),
					});
				},
				"string" => {
					traits.insert(ky.to_owned(),KratosTraitItem {
						typ: value,
						properties: None,
					});
				},
				_ => {
					println!("Type {} undefined",value.as_str());
				},
			}
		}
	};
	traits
}
pub fn set_kratos_user(mode: &str, data: HashMap::<String,String>,traits:  HashMap::<String,KratosTraitItem>) -> String {
  let mut user_data = String::from("{");
	for (ky,def) in traits {
		match def.typ.as_str() {
			"object" => {
	      if let Some(props) =  def.properties {
	        if mode == "json" {
						user_data += &format!("\"{}\": ",ky);
					  user_data += &"{";
					}
					for (it_ky,_it_def) in props {
						let key = format!("{}.{}",ky,it_ky);
					  if let Some(value) = data.get(&key) {
	            if mode == "json" {
					      user_data += &format!("\"{}\": \"{}\",",it_ky,value);
							} else {
					      user_data += &format!("\"{}.{}.{}\": \"{}\",",mode,ky,it_ky,value);
							}
						}	 
					}
	        if mode == "json" {
						user_data += &"}#,";
						user_data = user_data.replace(",}#","}");
					}
				}
			},
			_ => { 
				if let Some(value) = data.get(&ky) {
	        if mode == "json" {
					  user_data += &format!("\"{}\": \"{}\",",ky,value);
					} else {
					  user_data += &format!("\"{}.{}\": \"{}\",",mode,ky,value);
					}
				}
			},
		}
	}
	if let Some(passwd) = data.get(PASSWORD_FIELD) {
		user_data += &format!("\"{}\":\"{}\",",PASSWORD_FIELD,passwd);
		user_data += &format!("\"method\":\"{}\",",PASSWORD_FIELD);
	}
	user_data += &"}#";
	user_data = user_data.replace(",}#","}");
  user_data
}
pub fn set_kratos_login(data: HashMap::<String,String>) -> String {
  let mut user_data = String::from("{");
	for (ky,val) in data {
		user_data += &format!("\"{}\": \"{}\",",ky,val);
	}
  user_data += &format!("\"method\":\"{}\",",PASSWORD_FIELD);
	user_data += &"}#";
	user_data = user_data.replace(",}#","}");
  user_data
}
pub fn get_req_cli(cfg: SigninServer,cookie_store: &Arc<CookieStoreMutex>) -> Result<reqwest::Client> {
//pub fn get_req_cli(cfg: SigninServer) -> Result<reqwest::Client> {
	let mut req_cli = reqwest::Client::builder();
	if cfg.conn_timeout > 0 {
    req_cli = req_cli.connect_timeout(std::time::Duration::from_millis(cfg.conn_timeout));
	} 
	if cfg.timeout > 0 {
    req_cli = req_cli.timeout(std::time::Duration::from_millis(cfg.timeout));
	} 
	//match req_cli.cookie_provider(std::sync::Arc::clone(cookie_store)).build() {
	//match req_cli.cookie_store(true).build() {
	match req_cli.build() {
		Ok(reqcli) => Ok(reqcli),
		Err(e) => Err(anyhow!("Error get req client builder {}", &e)), 
	}
}
pub fn create_headers(id: &str,token: &str,csrf_token: &str, cookie: &str) -> Result<HeaderMap> {
	let mut headers = HeaderMap::new();
  headers.insert(USER_AGENT, HeaderValue::from_static("reqwest"));
  headers.insert(CONTENT_TYPE, HeaderValue::from_static("application/json"));
  headers.insert(ACCEPT, HeaderValue::from_static("application/json"));
	// if cookie != "" {
  //   headers.insert(COOKIE, HeaderValue::from_str(cookie)
	// 		.unwrap_or(HeaderValue::from_static(""))
	// 	);
  //   // headers.insert(SET_COOKIE, HeaderValue::from_str(cookie)
	// 	// 	.unwrap_or(HeaderValue::from_static(""))
	// 	// );
	// }
	if token != "" {
		headers.insert(AUTHORIZATION, HeaderValue::from_str(&format!("Bearer {}",token))
			.unwrap_or(HeaderValue::from_static(""))
		);
	}
	// if csrf_token != "" {
  //   headers.insert("X-CSRF-Token", 
	//     HeaderValue::from_str(csrf_token).unwrap_or(HeaderValue::from_static(""))
	//   );
	// }
	if id != "" {
    headers.insert("x-kratos-authenticated-identity-id", 
	    HeaderValue::from_str(id).unwrap_or(HeaderValue::from_static(""))
	  );
	}
	// dbg!("{:#?}",&headers);
	// dbg!("{:#?}",&headers);
	Ok(headers)
}
pub fn write_cookies(cfg: SigninServer,cookie_store: &Arc<CookieStoreMutex>) -> Result<()> {
  let mut writer = std::fs::File::create(cfg.cookies_path.to_owned())
      .map(std::io::BufWriter::new)?;
  match cookie_store.lock() {
		Ok(store) => {
      store.save_json(&mut writer).unwrap_or_else(|e|{
		    println!("Error save cookies to {}: {}",cfg.cookies_path,e);
			});
		},
		Err(e) => { println!("Error save cookies: {}",e);},
	};
	Ok(())
}
pub fn get_root_url(protocol: String, base: String, port: usize) -> String {
	format!("{}://{}:{}",protocol,base,port)
}
pub fn write_store_cookie(cfg: SigninServer, cookie_store: &Arc<CookieStoreMutex>) {
 let _ = write_cookies(cfg.to_owned(), cookie_store);
 match cookie_store.lock() {
		Ok(store) => {
			for c  in store.iter_any() {
			  let json_data = serde_json::to_string(&c).unwrap();
				println!("c: {:#?}", c);
				println!("{}", json_data);
			}
		},
		Err(e) => { println!("Error save cookies: {}",e);},
	};
}
pub async fn get_flow_csrf(token: &str,cfg: SigninServer, cookie_store: &Arc<CookieStoreMutex>, csrf_token: &str, cookie: &str, url: &str) -> Result<KratosReqId> {
// pub async fn get_flow_csrf(token: &str,cfg: SigninServer, csrf_token: &str, cookie: &str, url: &str) -> Result<KratosReqId> {
	let reqid: KratosReqId;
	let req_cli = get_req_cli(cfg.to_owned(), cookie_store)?;
	let headers = create_headers("",token,csrf_token,cookie)?;
  match req_cli
	  .get(url)
    // .bearer_auth(token.access_token().secret())
		.headers(headers)
    .send()
    .await {
			Ok(data) => {
				let flow: String;
				let csrf: String;
				let cookie: String;
				if let Some(ckie) = data.headers().get(SET_COOKIE) {
				  let str_cookie = &format!("{:?}",&ckie).replace("\"","");
					let (cookie_val,_) = str_cookie.split_once(";").unwrap_or_else(||("",""));
					cookie=cookie_val.to_owned();
				} else {	
					cookie = String::from("");
				}	
	      //dbg!("{:#?}",&data);
				if ! url.contains("logout") { // && data.status() == StatusCode::OK {
					match data.json::<KratosResponse>().await {
						Ok(response) => { 
							//  dbg!("{:#?}",&response);
							//let (_url_path,url_flow) = &response.ui.action.as_str().split_once("flow=").unwrap_or_else(||("",""));
							//flow=url_flow.to_string();
							flow=response.id;
							if let Some(req_csrf_token) = response.ui.nodes.iter().filter(|n| n.attributes.name.as_str() == "csrf_token").collect::<Vec<&KratosNode>>().into_iter().nth(0) {
								csrf=req_csrf_token.attributes.value.to_string();
							} else {
								csrf=String::from("");
							}
							reqid = KratosReqId { flow, csrf, cookie };
						},
						Err(e) => {
							return Err(anyhow!("Error get flow json: {}", &e));
						}
					};
				} else {
					match data.json::<serde_json::Value>().await {
						Ok(response) => { 
							dbg!("{:#?}",&response);
							//let (_url_path,url_flow) = &response.ui.action.as_str().split_once("flow=").unwrap_or_else(||("",""));
							//flow=url_flow.to_string();
							// flow=response.id;
							// if let Some(req_csrf_token) = response.ui.nodes.iter().filter(|n| n.attributes.name.as_str() == "csrf_token").collect::<Vec<&KratosNode>>().into_iter().nth(0) {
							// 	csrf=req_csrf_token.attributes.value.to_string();
							// } else {
							// 	csrf=String::from("");
							// }
							reqid = KratosReqId { flow: String::from(""), csrf: String::from(""), cookie: String::from("") };
						},
						Err(e) => {
							return Err(anyhow!("Error get flow json: {}", &e));
						}
					};
				}
			},
			Err(e) => {
				return Err(anyhow!("Error get flow {}", &e));
			},
	};
	Ok(reqid)
}
pub async fn on_kratos_user(token: &str,cfg: SigninServer, cookie_store: &Arc<CookieStoreMutex>, csrf_token: &str, cookie: &str, url: &str, data: String,) -> Result<OnKratosReq> {
// pub async fn on_kratos_user(token: &str, cfg: SigninServer, csrf_token: &str, cookie: &str,url: &str, data: String,) -> Result<OnKratosReq> {
	// let mut body_data = String::from("{");
	// // body_data += &format!("\"csrf_token\":\"{}\",#{}",csrf_token,data);
	// body_data += &format!("\"csrf_token\":\"{}\",#{}",csrf_token,data);
	// body_data = body_data.replace("#{","");
	let  body_data = String::from(data);
	println!("{}",&body_data);
	let json_data: serde_json::Value = serde_json::from_str(&body_data).unwrap_or_else(|e|{
    println!("Error json: {}",e);
		serde_json::Value::default()
	});
	let req_cli = get_req_cli(cfg.to_owned(), cookie_store)?;
	let headers = create_headers("",token,csrf_token,cookie)?;
	println!("URL: {}",&url);
	let status: StatusCode;
	let cookie: String;
	let session_resp: KratosSessionResponse;
	match req_cli
		.post(url)
		.headers(headers)
		.json(&json_data)
		.send()
		.await {
			Ok(res) => {
				if let Some(ckie) = res.headers().get(SET_COOKIE) {
				  let str_cookie = &format!("{:?}",&ckie).replace("\"","");
					dbg!("{:#?}",&ckie);
					let (cookie_val,_) = str_cookie.split_once(";").unwrap_or_else(||("",""));
					cookie=cookie_val.to_owned();
				} else {	
					cookie = String::from("");
				}
				status = res.status();
				if url.contains("login") || url.contains("whoami") {
				  // match res.json::<serde_json::Value>().await {
				  match res.json::<KratosSessionResponse>().await {
						Ok(data) => {
							// dbg!("{:#?}",&data);
					    // session_resp = KratosSessionResponse::default();
							session_resp = data;
						},
						Err(e) => {
							return Err(anyhow!("Error on kratos response session json  {}", &e));
						}
					}
				} else {
					session_resp = KratosSessionResponse::default();
				}
			},
			Err(e) => {
				println!("Error get :{}",e);
				return Err(anyhow!("Error on kratos  {}", &StatusCode::BAD_REQUEST));
			},
		// match response.status() {
		// reqwest::StatusCode::OK => {
		//     println!("Success! {:?}");
		// },
		// reqwest::StatusCode::UNAUTHORIZED => {
		//     println!("Need to grab a new token");
		// },
		// _ => {
		//     panic!("Uh oh! Something unexpected happened.");
		// },
	};
	Ok(OnKratosReq { status,cookie,session_resp })
}
pub async fn get_kratos_user(id: &str,token: &str,cfg: SigninServer,cookie_store: &Arc<CookieStoreMutex>, csrf_token: &str, cookie: &str, url: &str) -> Result<OnKratosReq> {
//pub async fn get_kratos_user(id: &str,token: &str,cfg: SigninServer, csrf_token: &str, cookie: &str, url: &str) -> Result<OnKratosReq> {
	// let reqid: KratosReqId;
	let req_cli = get_req_cli(cfg.to_owned(), cookie_store)?;
	let headers = create_headers(id,token,csrf_token,cookie)?;
	// dbg!("{:#?}",&headers);
	let status: StatusCode;
	let cookie = String::from("");
	let session_resp: KratosSessionResponse;
  match req_cli
	  .get(url)
		.headers(headers)
    .send()
    .await {
			Ok(res) => {
	      // dbg!("{:#?}",&res);
				status = res.status();
				match res.json::<serde_json::Value>().await {
				 //match res.json::<KratosSessionResponse>().await {
					Ok(data) => {
						// dbg!("{:#?}",&data);
				    session_resp = KratosSessionResponse::default();
				// if url.contains("logout") {
				// 	session_resp = KratosSessionResponse::default();
				// } else {
				// 	match res.json::<KratosSession>().await {
				// 		Ok(session) =>  session_resp = KratosSessionResponse {
				// 			session_token: token.to_owned(),
				// 			session,
					},
					Err(e) => {
						return Err(anyhow!("Error get kratos session json  {}", &e));
					}
				}
			},
			Err(e) => {
				println!("Error get :{}",e);
				return Err(anyhow!("Error get kratos user  {}", &StatusCode::BAD_REQUEST));
			},
	};
	Ok(OnKratosReq { status,cookie,session_resp })
}
pub async fn logout_kratos_user(id: &str,token: &str,cfg: SigninServer, cookie_store: &Arc<CookieStoreMutex>, csrf_token: &str, cookie: &str, url: &str) -> Result<OnKratosReq> {
	// let reqid: KratosReqId;
	let mut body_data = String::from("{");
	body_data += &format!("\"session_token\":\"{}\"",&token);
	// body_data += &format!("\"ory_kratos_session\":\"{}\"",&token);
	body_data += "}";
	println!("{}",&body_data);
	let json_data: serde_json::Value = serde_json::from_str(&body_data).unwrap_or_else(|e|{
    println!("Error json: {}",e);
		serde_json::Value::default()
	});
	let req_cli = get_req_cli(cfg.to_owned(), cookie_store)?;
	let headers = create_headers(id,token,csrf_token,cookie)?;
	// dbg!("{:#?}",&headers);
	let status: StatusCode;
	let cookie = String::from("");
	let session_resp: KratosSessionResponse;
  match req_cli
	  .delete(url)
	//  .get(url)
		.headers(headers)
		.json(&json_data)
    .send()
    .await {
			Ok(res) => {
	      dbg!("{:#?}",&res);
				status = res.status();
				if status == StatusCode::OK {
					match res.json::<serde_json::Value>().await {
					//match res.json::<KratosSessionResponse>().await {
						Ok(data) => {
							dbg!("{:#?}",&data);
							session_resp = KratosSessionResponse::default();
						},
						Err(e) => {
							return Err(anyhow!("Error parse logout kratos session json  {}", &e));
						},
					}
				} else {
					session_resp = KratosSessionResponse::default();
		   	}
//				if url.contains("logout") {
				// } else {
				// 	match res.json::<KratosSession>().await {
				// 		Ok(session) =>  session_resp = KratosSessionResponse {
				// 			session_token: token.to_owned(),
				// 			session,
				// 		},
				// 		Err(e) => {
				// 			return Err(anyhow!("Error get kratos session json  {}", &e));
				// 		}
				// 	};
				// }
			},
			Err(e) => {
				println!("Error get :{}",e);
				return Err(anyhow!("Error logout kratos user  {}", &StatusCode::BAD_REQUEST));
			},
	};
	Ok(OnKratosReq { status,cookie,session_resp })
}